{{  FullDuplexSerial4pnf version 2.1
  - Tracy Allen (TTA)   (c)22-Jan-2011, 2/12/2015   MIT license, see end of file for terms of use.  Extends existing terms of use.
  - based on Tim Moore's pcFullDuplexSerial4fc that can open up to 4 independent serial ports, using one cog for all 4
  - in turn from Chip Gracey fullDuplexSerial.
  - Also a thanks to Duane Degn and Wayne Meretsky (ksltd) and kuroneko for active suggestions.
  - Also based on TTA original fullDuplexSerial4port with flow control. But.............
  - This version (pnf) no longer supports flow control
  - This version has better coroutine management and addresses several other issues as noted below in the version notes.
  - This version does still support open and inverted baud modes on each channel
  - and individually configurable tx and rx buffers for all 4 ports, any size, set in CONstants section at compile time
  - Part of the buffer fits within the object's hub footprint. The object remains restartable without reload from EEPROM
  - Buffers are DAT type variables, therefore a single instance of the object can be accessed throughout a complex project.
  
  - 3/12/2018 corrected bug in DecPW, initialize sign~, otherwise positive numbers can print as negative.

}}
CON

  TRUMODE                       = %000000
  INVMODE                       = %110011

  NOMODE                        = %000000
  INVERTRX                      = %000001
  INVERTTX                      = %000010
  OCTX                          = %000100
  NOECHO                        = %001000
  INVERTCTS                     = %010000
  INVERTRTS                     = %100000

  PINNOTUSED                    = -1                    'tx/tx/cts/rts pin is not used
  DEFAULTTHRESHOLD              = 0                     ' zero defaults to 3/4 of buffer length

  RADIX                         = "."                   ' for decimal point display, method DecDP(....)
  FILLER                        = 32                    ' for decimal numeric display, field width filler char for right justified number

  TX_MAX_WAIT                   = 100                   ' wait only 100milliseconds max for the tx port to clear for the next character
                                                        ' not implemented in the code.  only important with handshake enabled.
  NOECHO_WAIT                   = 1                     ' wait only one millisecond for an echo from tx in NPECHO mode (for open baudmode to one wire)

  BAUD1200                      = 1200
  BAUD2400                      = 2400
  BAUD4800                      = 4800
  BAUD9600                      = 9600
  BAUD19200                     = 19200
  BAUD38400                     = 38400
  BAUD57600                     = 57600
  BAUD115200                    = 115200

' The following constants declare the sizes of the rx and tx buffers.
' Enter in the needed size in bytes for each rx and tx buffer
' These values can be any size within available memory. They do not have to be a power of two.
' Unused buffers can be reduced to 4 bytes.
  RX_SIZE0                      = 60  ' receive buffer allocations
  RX_SIZE1                      = 60
  RX_SIZE2                      = 60
  RX_SIZE3                      = 60

  TX_SIZE0                      = 32 ' transmit buffer allocations
  TX_SIZE1                      = 32
  TX_SIZE2                      = 32
  TX_SIZE3                      = 32   '


  RXTX_BUFSIZE                  = (TX_SIZE0 + TX_SIZE1 + TX_SIZE2 + TX_SIZE3 + RX_SIZE0 + RX_SIZE1 + RX_SIZE2 + RX_SIZE3)
                                 ' total buffer footprint in bytes
                                 ' as of this version 244 bytes are available for buffers within the hub footprint of the object
                                 ' the final instruction in this program allocates additional buffer space beyond that if necessary
                                 ' to accomodate all of the buffers.
                                 ' if the sum totals to 244, then the buffers exactly fit within the object footprint.
                                 ' The allocations of 50,24,24,24,50,24,24,24 meet that criterion.


PUB Init
''Always call init before adding ports
  Stop
  bytefill(@startfill, 0, (@endfill-@startfill))        ' initialize head/tails,port info and hub buffer pointers
  return @rxsize                                        ' TTA returns pointer to data structure, buffer sizes.  ports

PUB AddPort(port,rxpin,txpin,mode,baudrate)
'' Call AddPort to define each port, names refer to Prop as DTE
'' port 0-3 port index of which serial port
'' rx/tx/cts/rts pin number                             XXX#PINNOTUSED if not used
'' rx=data input
'' tx=data output
'' mode bit 0 = invert rx                              XXX#INVERTRX    invert means stop bits are low, start bits high.
'' mode bit 1 = invert tx                               XXX#INVERTTX    true means stop high, start low
''    non-inverted with mode bit = 0, start bit is low, stop bits are high
''    inverted with mode bit = 1, start bit is high, stop bits are low
'' mode bit 2 = open-drain/source tx                 XXX#OCTX
''    if mode is 4 or 5 tx open non-inverted, then a pullup resistor is needed to Vdd, pulls up to the state of the stop bits.
''    if mode is 6 or 7 tx open inverted, then a pulldown resistor is needed to Vss.  Stop bits are Vss
''     returns the size of the buffers as one long, rx buffer size in high word, rx buffer size in low word
'' mode bit 3 = ignore tx echo on rx                    XXX#NOECHO
'' baudrate
  if cog OR (port > 3)
    abort
  result := port*2
  ports := ports | (((rxpin > -1) & 1) << result) | (((txpin > -1) & 1)<< (result+1))
  ' ports := ports + |< port                              ' tracks which ports have been opened
  if rxpin <> -1
    long[@rxmask][port] := |< rxpin
  if txpin <> -1
    long[@txmask][port] := |< txpin
  long[@rxtx_mode][port] := mode
  long[@bit_ticks][port] := (clkfreq / baudrate)
  long[@bit4_ticks][port] := long[@bit_ticks][port] >> 2
  return rxSize[port]<<16 + txSize[port]   ' return buffer sizes as set up in the CONstants


PUB Start : okay
'' Call start to start cog
'' Start serial driver - starts a cog
'' returns false if no cog available
''
'' tx buffers will start within the object footprint, overlaying certain locations that were initialized in spin
'' for  use within the cog but are not needed by spin thereafter and are not needed for object restart.

  txbuff_tail_ptr := txbuff_ptr  := @buffers                   ' all buffers are calculated as offsets from this address.
  txbuff_tail_ptr1 := txbuff_ptr1 := txbuff_ptr + txsize       ' base addresses of the corresponding port buffer.
  txbuff_tail_ptr2 := txbuff_ptr2 := txbuff_ptr1 + txsize1
  txbuff_tail_ptr3 := txbuff_ptr3 := txbuff_ptr2 + txsize2
  rxbuff_head_ptr := rxbuff_ptr  := txbuff_ptr3 + txsize3      ' rx buffers follow immediately after the tx buffers, by size
  rxbuff_head_ptr1 := rxbuff_ptr1 := rxbuff_ptr + rxsize
  rxbuff_head_ptr2 := rxbuff_ptr2 :=  rxbuff_ptr1 + rxsize1
  rxbuff_head_ptr3 := rxbuff_ptr3 :=  rxbuff_ptr2 + rxsize2
                                                               ' note that txbuff_ptr ... rxbuff_ptr3 are the base addresses fixed
                                                               ' in memory for use by both spin and pasm
                                                               ' while txbuff_tail_ptr ... rxbuff_head_ptr3 are dynamic addresses used only by pasm
                                                               ' and here initialized to point to the start of the buffers.
                                                               ' the rx buffer #3 comes last, up through address @endfill
  rx_head_ptr  := @rx_head                                     '  addresses of the head and tail counts are passed to the cog
  rx_head_ptr1 := @rx_head1                                    ' if that is confusing, take heart.   These are pointers to pointers to pointers
  rx_head_ptr2 := @rx_head2
  rx_head_ptr3 := @rx_head3
  rx_tail_ptr  := @rx_tail
  rx_tail_ptr1 := @rx_tail1
  rx_tail_ptr2 := @rx_tail2
  rx_tail_ptr3 := @rx_tail3
  tx_head_ptr  := @tx_head
  tx_head_ptr1 := @tx_head1
  tx_head_ptr2 := @tx_head2
  tx_head_ptr3 := @tx_head3
  tx_tail_ptr  := @tx_tail
  tx_tail_ptr1 := @tx_tail1
  tx_tail_ptr2 := @tx_tail2
  tx_tail_ptr3 := @tx_tail3
  ticksPerMs := clkfreq/1000
  okay := cog := cognew(@entry, @rx_head) + 1
  waitcnt(clkfreq/8+cnt)                                       ' allows time for cog to start

PUB Stop
'' Stop serial driver - frees a cog
  ports := 0
  if cog
    cogstop(cog~ - 1)

PUB GetCogID
  return cog -1

PUB GetPorts                                                   ' ports variable is 8 bits, 1 in bit position for each active rx and tx
  return ports

PUB Rxflush(port)
'' Flush receive buffer, here until empty.
  repeat while rxcheck(port) <> -1

PUB RxHowFull(port)    ' (TTA) added method
'' returns number of chars in rx buffer
  return ((rx_head[port] - rx_tail[port]) + rxsize[port]) // rxsize[port]
'   rx_head and rx_tail are values in the range 0=< ... < RX_BUFSIZE

{PUB RxChunk(port, bufferPntr, Nmax) | head, tail, size, base, chunk, todo, rxN
' This has not been fully tested, idea is to receive faster double buffering.
  head := rx_head[port]
  tail := rx_tail[port]
  size := rxsize[port]
  base := rxbuff_ptr[port]                      ' Base hub addresses of the receive buffer
  if (rxN := (head - tail + size) // size)
  todo := rxN <# Nmax
    chunk := (size - tail) <# todo
    bytemove(bufferPntr, base+rx_tail, chunk)
    bytemove(bufferPntr+chunk, base, todo-chunk)
  rx_tail[port] := (head + todo) // size   ' adjust the tail pointer
  return todo               ' return how many have been transferred
}
PUB RxCheck(port) : rxbyte
'' Check if byte received (never waits)
'' returns -1 if no byte received, $00..$FF if byte
'' note that inverted character mode is now handled in the pasm
  rxbyte--                                                     ' initially -1 for no byte received
  if rx_tail[port] <> rx_head[port]
    rxbyte := byte[rxbuff_ptr[port] + rx_tail[port]]
    rx_tail[port] := (rx_tail[port] + 1) // rxsize[port]

PUB RxTime(port,ms) : rxbyte | t
'' Wait ms milliseconds for a byte to be received
'' returns -1 if no byte received, $00..$FF if byte
'' if ms is negative, waits forever (like rx[port])
'' oneMillisecond := clkfreq/1000 , set as ticksperms in Start method
'' note this returns rapidly if a byte is already available at entry (compared with original method)
  t := cnt
  repeat until (rxbyte := rxcheck(port)) => 0
    if ((cnt - t) > ticksPerMs * ms) AND (ms => 0)
       quit

PUB Rx(port) : rxbyte
'' Receive byte (may wait for byte)
'' returns $00..$FF
'' for bug protection, checks to see if the port is actually open
  if |<(port<<1) & ports
    repeat until (rxbyte := rxcheck(port)) => 0


PUB Tx(port,txbyte)
'' Send byte (may wait for room in buffer)
'  will add data to buffer only if port is enabled for tx, bug protection
'  Use the packet method below to send longer strings.
  if |<(port<<1+1) & ports
    repeat until (tx_tail[port] <> (tx_head[port] + 1) // txsize[port])  ' wait for space in buffer
    byte[txbuff_ptr[port]+tx_head[port]] := txbyte
    tx_head[port] := (tx_head[port] + 1) // txsize[port]

    if rxtx_mode[port] & NOECHO
      rxtime(port,NOECHO_WAIT)              ' waits this long in milliseconds if  NOECHO mode is specified

PUB Char(port,txByte)   ' compatibility
  tx(port, txByte)

PUB Chars(port, txByte, reps)  ' repeat a character
  repeat reps
    tx(port, txByte)

PUB TxFlush(port)
'  if  this port is open for tx ' was if port>3 abort.  Not so good.
  if |<(port<<1+1) & ports
    repeat until (long[@tx_tail][port] == long[@tx_head][port])

PUB Packet(port, bytePointer, byteCount) | free, todo, chunk, head, tail, base, size
  ' This is adapted from Wayne Meretsky's serial object
  ' Like method StrN, but much faster for transfers in Spin with double buffering
  if |<(port<<1+1) & ports   ' don't try it unless this port is enabled for tx
    size := txSize[port]
    base := txbuff_ptr[port]                      ' Base hub addresses of the transmit buffer
    repeat while (byteCount <> 0)
      head := tx_head[port]
      tail := tx_tail[port]
      free := tail - head                                        ' free bytes in buffer
      if (free =< 0)                                             ' note this has to be =< so that free-1 won't end up as -1
        free += size
      todo := (free - 1) <# byteCount                            ' number of bytes to move now
      chunk := (size - head) <# todo                            ' bytes to the end of the buffer
      bytemove (base + head, bytePointer,  chunk)               ' move bytes up to the end of the buffer
      bytemove (base, bytePointer + chunk, todo - chunk)        ' Fill the rest at top of buffer; often none
      tx_head[port] := (head + todo) // size                    ' adjust the pointer
      byteCount -= todo                                         ' remaining number of bytes
      bytePointer += todo                                       ' from this point in the buffer

PUB Str(port,stringptr)
'' Send zstring
  StrN(port,stringptr,strsize(stringptr))

PUB StrN(port,stringptr,nchar)
'' Send counted string
  repeat nchar
    tx(port,byte[stringptr++])

PUB Hex(port, value, digits)
'' Print a hexadecimal number
  value <<= (8 - digits) << 2
  repeat digits
    tx(port, lookupz((value <-= 4) & $F : "0".."9", "A".."F"))

PUB Bin(port, value, digits)
'' Print a binary number
  value <<= 32 - digits
  repeat digits
    tx(port, (value <-= 1) & 1 + "0")

PUB Dec(port,value)
' standard decimal value, no decimal point, only as many digits as required
  DecPW(port,value,0,0)

PUB DecP(port, value, places)
' with a decimal point, places is the number of digits to the right of the radix
  DecPW(port, value, places, 0)

PUB DecPW(port, value, places, width) | sign, idx, box[4]
' A decimal value with optional radix point is right justified in a field padded on the left with spaces.
' Then printed using the counted string method, strN
  sign~
  width <#= 15                                          ' limit width of field to 15 chars
  result := (value == NEGX)                               'if NEGX  result=-1 else result=0 for later action
  if value < 0
    value := ||(value + result)                          ' absoulute value, negx becomes posx 2147483647
    sign~~                                              ' will put "-" later in the buffer
  bytefill(@box, 32, 15)                                ' fill the buffer with spaces
  byte[@box+15] := 0                                    ' zstring
  repeat idx from 0 to 14
    if idx == places and places > 0                     ' insert the radix if any at its proper place
      byte[@box+14-idx] := RADIX
    else
      byte[@box+14-idx] := value//10 + result*(idx == 0) + "0"        ' peel decimal digits off at the least significant end.                      '
      value/=10                                                  ' put in buffer, and handle NEGX by changing 1's digit 7>>8
    if (value == 0) and (idx => places)                  ' have both run out of digits and have come to left of radix position
      quit    
  if (idx++ == places) and (places > 0)
    byte[@box+14-idx++] := "0"                          ' leading zero in front of decimal point
  if sign
    byte[@box+14-idx++] := "-"                          ' put the "-" in the buffer in front of the number
  if idx > width                                        ' leading spaces to fill width, but always the entire number
    width := idx                                        ' comment this out if you want to trim the number to width, or fill with warning
  result := @box+15-width                                    ' the method returns with a pointer to the value string, in case it needs to be reprinted.
  StrN(port, result, width)                              ' print the string
  return result                                          ' also return the pointer to the string (in stack space, consume or copy immediately)


DAT
'***********************************
'* Assembly language serial driver *
'***********************************
'
                        org 0
'
' Entry
'                   
'The initialization code checks the modes and modifies the rx/tx code for that mode
'e.g. the if condition for rx checking for a start bit will be inverted if mode INVERTRX
'is it, similar for other mode flags
'Similar with the co-routine variables. If an rx or tx pin is not configured
'the jmpret for the routine that handles that pin is modified so the routine is never called
'We start with port 3 and work down to ports because we will be updating the co-routine pointers
'and the order matters. e.g. we can update txcode3 and then update rxcode3 based on txcode3.
' (TTA): coroutine patch was not working in the way originally described.
' (TTA): changed the way the coroutine vectors are patched, now correctly

' The following 8 locations are skipped at entry due to if_never.
' The mov instruction and the destination address are here only for syntax.
' the important thing are the source field
' primed to contain the start address of each port routine.
' When jmpret instructions are executed, the source adresses here are used for jumps
' And new source addresses will be written in the process.
entry
rxcode  if_never        mov     rxcode,#receive        ' set source fields to initial entry points
txcode  if_never        mov     txcode,#transmit       ' these will be modified by the jmpret mechanism
rxcode1 if_never        mov     rxcode1,#receive1       ' never executed,only the source fields matter as vectors to coroutine code
txcode1 if_never        mov     txcode1,#transmit1      ' why here?  It makes the numbering easy, indirect jumps via location 0,1,2,3,4,5,6,7
rxcode2 if_never        mov     rxcode2,#receive2
txcode2 if_never        mov     txcode2,#transmit2
rxcode3 if_never        mov     rxcode3,#receive3
txcode3 if_never        mov     txcode3,#transmit3
brcode  if_never        mov     brcode,#bridge

' ---- INITIALIZATIONS ==============================================================================
' find lowest active port and seed it to the following bubble up initialization
' method AddPort has filled variable "ports" with a 1 in each bit position for an active rx or tx, 8 bits total for 4 ports
bridgePort              shr     ports, #1   wc,wz       '
        if_nc           add     t1, #1                  ' this loop finds the lowest numbered active port to be included in the jmpret cycle
        if_nc           jmp     #bridgePort            '  note it will hang here if no port is enabled
        if_z            movs    bridge,t1               ' if ports is zero at this point, then there is only one port active
        if_z            mov     t1, #brcode             '  so the loop needs to include bridge. A jmpret can't reference itself.

' ---- port 3 initialization -------------------------------------------------------------
                        test    rxtx_mode3,#OCTX     wz       'init tx pin according to mode
                        test    rxtx_mode3,#INVERTTX wc
        if_z_ne_c       or      outa,txmask3                  ' for open baudmode
        if_z            or      dira,txmask3                  ' for output driven high and low
        if_z_and_nc     andn    txout3,domuxnc                ' patch muxc to muxnc
        if_nz           movd    txout3,#dira                  ' change destination from outa to dira
                                                              ' patch rx wait for start bit depending on invert
                        test    rxtx_mode3,#INVERTRX wz       ' will start bit on rx pin be low or high level?
        if_nz           xor     start3,doifc2ifnc             ' if_c jmp to if_nc, start bit it high, not low
        if_nz           or      stop3-1, never2always         ' include code to invert the byte if necessary
        if_nz           xor     stop3,doifc2ifnc   ' inverted mode stop bit is low

                        or      txmask3,#0      wz,nr         ' nz if tx3 is active in the loop
                        movs    transmit3, t1                 ' its vectors will point back to the lowest numbered active port vector or to brcode if only one active
                        movs    tx3jr1, t1
                        movs    tx3jr2, t1
        if_nz           mov     t1, #txcode3                   ' if tx3 is used, t1 will now point to txcode3 to include it in the path, else t1 still points to the lowest active port
                        or      rxmask3,#0      wz,nr          '  ditto for rx routine
                        movs    receive3, t1                   ' if used, vector through either the txcode3 or through the lowest active port
                        movs    rx3jr1, t1
                        movs    rx3jr2, t1
        if_nz           mov     t1, #rxcode3                   ' if rx3 is used, t1 will now point to rxcode3 so rx3 will be included in the path

' ---- port 2 initialization -------------------------------------------------------------
                        test    rxtx_mode2,#OCTX wz            ' same comments as port 3 initialization
                        test    rxtx_mode2,#INVERTTX wc
        if_z_ne_c       or      outa,txmask2
        if_z            or      dira,txmask2
        if_z_and_nc     andn    txout2,domuxnc
        if_nz           movd    txout2,#dira
                        test    rxtx_mode2,#INVERTRX wz
        if_nz           xor     start2,doifc2ifnc
        if_nz           or      stop2-1, never2always
        if_nz           xor     stop2,doifc2ifnc   ' inverted mode stop bit is low

                        or      txmask2,#0      wz,nr
                        movs    transmit2, t1
                        movs    tx2jr1, t1
                        movs    tx2jr2, t1
        if_nz           mov     t1, #txcode2                   ' note the value of t1 is bubbling up the acive port sequency
                        or      rxmask2,#0      wz,nr
                        movs    receive2, t1
                        movs    rx2jr1, t1
                        movs    rx2jr2, t1
        if_nz           mov     t1, #rxcode2                   ' so that inactive ports can be shut out of the main loop

' ---- port 1 initialization -------------------------------------------------------------
                        test    rxtx_mode1,#OCTX wz
                        test    rxtx_mode1,#INVERTTX wc
        if_z_ne_c       or      outa,txmask1
        if_z            or      dira,txmask1
        if_z_and_nc     andn    txout1,domuxnc
        if_nz           movd    txout1,#dira
                        test    rxtx_mode1,#INVERTRX wz
        if_nz           xor     start1,doifc2ifnc
        if_nz           or      stop1-1, never2always
        if_nz           xor     stop1,doifc2ifnc   ' inverted mode stop bit is low

                        or      txmask1,#0      wz,nr
                        movs    transmit1, t1
                        movs    tx1jr1, t1
                        movs    tx1jr2, t1
        if_nz           mov     t1, #txcode1
                        or      rxmask1,#0      wz,nr
                        movs    receive1, t1
                        movs    rx1jr1, t1
                        movs    rx1jr2, t1
        if_nz           mov     t1, #rxcode1

' ---- port 0 initialization -------------------------------------------------------------
                        test    rxtx_mode,#OCTX wz
                        test    rxtx_mode,#INVERTTX wc
        if_z_ne_c       or      outa,txmask
        if_z            or      dira,txmask
        if_z_and_nc     andn    txout0,domuxnc
        if_nz           movd    txout0,#dira
                        test    rxtx_mode,#INVERTRX wz   ' z=0 means stop
        if_nz           xor     start0,doifc2ifnc
        if_nz           or      stop0-1, never2always
        if_nz           xor     stop0,doifc2ifnc   ' inverted mode stop bit is low

                        or      txmask,#0      wz,nr
                        movs    transmit, t1
                        movs    tx0jr1, t1
                        movs    tx0jr2, t1
        if_nz           mov     t1, #txcode
                        or      rxmask,#0      wz,nr
                        movs    receive, t1
                        movs    rx0jr1, t1
                        movs    rx0jr2, t1
        if_nz            mov     t1, #rxcode


' MAIN LOOP  =======================================================================================
' Receive0 -------------------------------------------------------------------------------------
receive                 jmpret  rxcode,txcode                  ' note that the source field may be modified during initialization
                        test    rxmask,ina      wc
start0  if_c            jmp     #receive                       ' may be patched fom if_c to if_nc for low start bit
                        mov     rxbits,#9                      ' ready to receive byte
                        mov     rxcnt,bit4_ticks               ' 1/4 bit
                        add     rxcnt,cnt                          
bit0                    add     rxcnt,bit_ticks                ' 1 bit period

rx0jr1                  jmpret  rxcode,txcode                  ' run a chuck of transmit code, then return
                        mov     t1,rxcnt                       ' check if bit receive period done
                        sub     t1,cnt
                        cmps    t1,#0           wc
        if_nc           jmp     #rx0jr1
                        test    rxmask,ina      wc             ' receive bit on rx pin,  (stop bit noninverted is high)
                        rcr     rxdata,#1
                        djnz    rxbits,#bit0                   ' get remaining bits
        if_never        xor     rxdata, neg1                   ' patch this to if_always do the byte inversion
stop0   if_nc            jmp     #receive                       ' abort if no stop bit   non-inverted stop bit is high, c=1, inverted c=0

rx0jr2                  jmpret  rxcode,txcode                  ' run a chunk of transmit code, then return
                        shr     rxdata,#32-9                   ' justify and trim received byte
                        wrbyte  rxdata,rxbuff_head_ptr'{7-22}  ' 1wr
                        add     rx_head,#1
                        cmpsub  rx_head,rxsize                 ' allows non-binary buffer size
                        wrlong  rx_head,rx_head_ptr   '{8}     ' 2wr
                        mov     rxbuff_head_ptr,rxbuff_ptr      'calculate next byte head_ptr
                        add     rxbuff_head_ptr,rx_head
                        jmp     #receive                        'byte done, receive next byte
'
' Receive1 -------------------------------------------------------------------------------------
'
receive1                jmpret  rxcode1,txcode1
                        test    rxmask1,ina     wc
start1  if_c            jmp     #receive1
                        mov     rxbits1,#9
                        mov     rxcnt1,bit4_ticks1
                        add     rxcnt1,cnt                          
bit1                    add     rxcnt1,bit_ticks1
                        
rx1jr1                  jmpret  rxcode1,txcode1
                        mov     t1,rxcnt1
                        sub     t1,cnt
                        cmps    t1,#0           wc
        if_nc           jmp     #rx1jr1
                        test    rxmask1,ina     wc
                        rcr     rxdata1,#1
                        djnz    rxbits1,#bit1
        if_never        xor     rxdata1, neg1
stop1   if_nc            jmp     #receive1

rx1jr2                  jmpret  rxcode1,txcode1
                        shr     rxdata1,#32-9

                        wrbyte  rxdata1,rxbuff_head_ptr1
                        add     rx_head1,#1
                        cmpsub  rx_head1,rxsize1
                        wrlong  rx_head1,rx_head_ptr1
                        mov     rxbuff_head_ptr1,rxbuff_ptr1
                        add     rxbuff_head_ptr1,rx_head1
                        jmp     #receive1
'
' Receive2 -------------------------------------------------------------------------------------
'
receive2                jmpret  rxcode2,txcode2
                        test    rxmask2,ina     wc
start2 if_c             jmp     #receive2
                        mov     rxbits2,#9
                        mov     rxcnt2,bit4_ticks2
                        add     rxcnt2,cnt
bit2                    add     rxcnt2,bit_ticks2
                        
rx2jr1                  jmpret  rxcode2,txcode2
                        mov     t1,rxcnt2
                        sub     t1,cnt
                        cmps    t1,#0           wc
        if_nc           jmp     #rx2jr1
                        test    rxmask2,ina     wc,wz
                        rcr     rxdata2,#1
                        djnz    rxbits2,#bit2
        if_never        xor     rxdata2, neg1
stop2   if_nc           jmp     #receive2

rx2jr2                  jmpret  rxcode2,txcode2
                        shr     rxdata2,#32-9
                        wrbyte  rxdata2,rxbuff_head_ptr2
                        add     rx_head2,#1
                        cmpsub  rx_head2,rxsize2
                        wrlong  rx_head2,rx_head_ptr2
                        mov     rxbuff_head_ptr2,rxbuff_ptr2
                        add     rxbuff_head_ptr2,rx_head2
                        jmp     #receive2
'
' Receive3 -------------------------------------------------------------------------------------
'
receive3                jmpret  rxcode3,txcode3
                        test    rxmask3,ina     wc
start3 if_c             jmp     #receive3
                        mov     rxbits3,#9
                        mov     rxcnt3,bit4_ticks3
                        add     rxcnt3,cnt
bit3                    add     rxcnt3,bit_ticks3
                        
rx3jr1                  jmpret  rxcode3,txcode3
                        mov     t1,rxcnt3
                        sub     t1,cnt
                        cmps    t1,#0           wc
        if_nc           jmp     #rx3jr1
                        test    rxmask3,ina     wc
                        rcr     rxdata3,#1
                        djnz    rxbits3,#bit3
        if_never        xor     rxdata3, neg1
stop3   if_nc            jmp     #receive3

rx3jr2                  jmpret  rxcode3,txcode3
                        shr     rxdata3,#32-9
                        wrbyte  rxdata3,rxbuff_head_ptr3
                        add     rx_head3,#1
                        cmpsub  rx_head3,rxsize3
                        wrlong  rx_head3,rx_head_ptr3
                        mov     rxbuff_head_ptr3,rxbuff_ptr3
                        add     rxbuff_head_ptr3,rx_head3
                        jmp     #receive3
'
' TRANSMIT =======================================================================================
'
transmit                jmpret  txcode,rxcode1        'run a chunk of receive code, then return
                        rdlong  t1,tx_head_ptr        '{7-22} - head[0]
                        cmp     t1,tx_tail      wz    'tail[0]
tx0l1   if_z            jmp     #transmit             '
                        rdbyte  txdata,txbuff_tail_ptr '{8}
                        add     tx_tail,#1
                        cmpsub     tx_tail,txsize    wz   ' (TTA) for individually sized buffers, will zero at rollover
                        wrlong  tx_tail,tx_tail_ptr    '{8}  
        if_z            mov     txbuff_tail_ptr,txbuff_ptr 'reset tail_ptr if we wrapped
        if_nz           add     txbuff_tail_ptr,#1    'otherwise add 1

tx0jr1                  jmpret  txcode,rxcode1
                        shl     txdata,#1
                        or      txdata,txbitor1       'ready byte to transmit
                        mov     txbits,#10
                        mov     txcnt,cnt
txbit                   shr     txdata,#1       wc
txout0                  muxnc    outa,txmask           'maybe patched to muxnc dira,txmask
                        add     txcnt,bit_ticks       'ready next cnt

tx0jr2                  jmpret  txcode,rxcode1        'run a chunk of receive code, then return
                        mov     t1,txcnt              'check if bit transmit period done
                        sub     t1,cnt
                        cmps    t1,#0           wc
        if_nc           jmp     #tx0jr2
                        djnz    txbits,#txbit         'another bit to transmit?
txjmp0                  jmp     #transmit                'byte done, transmit next byte
'
' Transmit1 -------------------------------------------------------------------------------------
'
transmit1               jmpret  txcode1,rxcode2
                        rdlong  t1,tx_head_ptr1
                        cmp     t1,tx_tail1     wz
tx1l1   if_z            jmp     #transmit1            '
                        rdbyte  txdata1,txbuff_tail_ptr1
                        add     tx_tail1,#1
                        cmpsub     tx_tail1,txsize1   wz
                        wrlong  tx_tail1,tx_tail_ptr1
        if_z            mov     txbuff_tail_ptr1,txbuff_ptr1
        if_nz           add     txbuff_tail_ptr1,#1

tx1jr1                  jmpret  txcode1,rxcode2
                        shl     txdata1,#1
                        or      txdata1,txbitor1
                        mov     txbits1,#10
                        mov     txcnt1,cnt
txbit1                  shr     txdata1,#1      wc
txout1                  muxnc    outa,txmask1
                        add     txcnt1,bit_ticks1

tx1jr2                 jmpret  txcode1,rxcode2
                        mov     t1,txcnt1
                        sub     t1,cnt
                        cmps    t1,#0           wc
        if_nc           jmp     #tx1jr2
                        djnz    txbits1,#txbit1
txjmp1                  jmp     #transmit1
'
' Transmit2 -------------------------------------------------------------------------------------
'
transmit2               jmpret  txcode2,rxcode3
                        rdlong  t1,tx_head_ptr2
                        cmp     t1,tx_tail2     wz
tx2l1   if_z            jmp     #transmit2            '

                        rdbyte  txdata2,txbuff_tail_ptr2
                        add     tx_tail2,#1
                        cmpsub     tx_tail2,txsize2   wz
                        wrlong  tx_tail2,tx_tail_ptr2
        if_z            mov     txbuff_tail_ptr2,txbuff_ptr2
        if_nz           add     txbuff_tail_ptr2,#1

tx2jr1                  jmpret  txcode2,rxcode3
                        shl     txdata2,#1
                        or      txdata2,txbitor1
                        mov     txbits2,#10
                        mov     txcnt2,cnt
txbit2                  shr     txdata2,#1      wc
txout2                  muxnc    outa,txmask2
                        add     txcnt2,bit_ticks2

tx2jr2                  jmpret  txcode2,rxcode3
                        mov     t1,txcnt2
                        sub     t1,cnt
                        cmps    t1,#0           wc
        if_nc           jmp     #tx2jr2
                        djnz    txbits2,#txbit2
txjmp2                  jmp     #transmit2
'
' Transmit3 -------------------------------------------------------------------------------------
'
transmit3               jmpret  txcode3,bridge                 ' note that the source field may be modified during initialization
                        rdlong  t1,tx_head_ptr3
                        cmp     t1,tx_tail3     wz
tx3l1   if_z            jmp     #transmit3
                        rdbyte  txdata3,txbuff_tail_ptr3
                        add     tx_tail3,#1
                        cmpsub     tx_tail3,txsize3   wz
                        wrlong  tx_tail3,tx_tail_ptr3
        if_z            mov     txbuff_tail_ptr3,txbuff_ptr3 'reset tail_ptr if we wrapped
        if_nz           add     txbuff_tail_ptr3,#1   'otherwise add 1

tx3jr1                  jmpret  txcode3,bridge                 ' source field may be modified
                        shl     txdata3,#1
                        or      txdata3,txbitor1
                        mov     txbits3,#10
                        mov     txcnt3,cnt
txbit3                  shr     txdata3,#1      wc
txout3                  muxnc    outa,txmask3
                        add     txcnt3,bit_ticks3

tx3jr2                  jmpret  txcode3,bridge                 ' source field may be modified
                        mov     t1,txcnt3
                        sub     t1,cnt
                        cmps    t1,#0           wc
        if_nc           jmp     #tx3jr2
                        djnz    txbits3,#txbit3
txjmp3                  jmp     #transmit3

' bridge This only executes if only one other coroutine is enabled.  A coroutine cannot jmpret to itself -------------------------

bridge                  jmp rxcode                              ' simplified to jmp rxcode per suggestion by Kuroneko
                        '                                       ' this coroutine is only included when a single rx or tx is enabled
                                                                ' a jmpret cannot jmpret to itself due to modification of the destination field.

'
'The following are constants used by pasm for patching the code, depending on options required
' We won't overright the hub copies, in case the object needs to be restarted.
doifc2ifnc              long      $003c0000                    ' patch condition if_c to if_nc using xor
domuxnc                 long      $04000000                    ' patch muxc to muxnc using or
txbitor1                long      $fffffe00                    ' overlay for one start bit, many stop bits
never2always            long      $003c0000                    ' OR with a if_never to make it if_always
neg1                    long      $ffffffff

' Buffer sizes for rx and tx initialized from CONstants and used by both spin and pasm

rxsize                  long      RX_SIZE0                     ' (TTA) size of the rx and tx buffers is available to pasm and spin
rxsize1                 long      RX_SIZE1                     ' these values are transfered from the declared CONstants
rxsize2                 long      RX_SIZE2                     ' at startup, individually configurable
rxsize3                 long      RX_SIZE3
txsize                  long      TX_SIZE0
txsize1                 long      TX_SIZE1
txsize2                 long      TX_SIZE2
txsize3                 long      TX_SIZE3

bit_ticks             long      0                              ' clock ticks per bit, here so baud rates are accessible from spin
bit_ticks1            long      0
bit_ticks2            long      0
bit_ticks3            long      0
ticksPerMs            long      0                       ' at Start, this is the number of clock ticks in one millisecond.


' Object memory from here to the end is zeroed in the init/stop method ---------------'
' Some locations within the next set of values, after being initialized to zero, are then filled with alternative options
' That are accessed from both spin and pasm
' Dont Change the order of these initialized variables within port groups of 4 without modifying
' the code to match - both spin and assembler

startfill
ports                 long      0               ' bitmapped for the ports enabled.
cog                   long      0                   'cog flag/id
rxtx_mode             long      0             ' mode setting from values passed in by addport
rxtx_mode1            long      0             '
rxtx_mode2            long      0
rxtx_mode3            long      0
rx_head               long      0             ' rx head pointer, from 0 to size of rx buffer, used in spin and pasm
rx_head1              long      0             ' data is enqueued to this offset above base, rxbuff_ptr
rx_head2              long      0
rx_head3              long      0
rx_tail               long      0             ' rx tail pointer, ditto, zero to size of rx buffer
rx_tail1              long      0             ' data is dequeued from this offset above base, rxbuff_ptr
rx_tail2              long      0
rx_tail3              long      0
tx_head               long      0             ' tx head pointer, , from 0 to size of tx buffer, used in spin and pasm
tx_head1              long      0             ' data is enqueued to this offset above base, txbuff_ptr
tx_head2              long      0
tx_head3              long      0
tx_tail               long      0             ' tx tail pointer, ditto, zero to size of rx buffer
tx_tail1              long      0             ' data is transmitted from this offset above base, txbuff_ptr
tx_tail2              long      0
tx_tail3              long      0
rxbuff_ptr            long      0             ' These are the base hub addresses of the receive buffers
rxbuff_ptr1           long      0             ' initialized in spin, referenced in pasm and spin
rxbuff_ptr2           long      0             ' these buffers and sizes are individually configurable
rxbuff_ptr3           long      0
txbuff_ptr            long      0             ' These are the base hub addresses of the transmit buffers
txbuff_ptr1           long      0
txbuff_ptr2           long      0
txbuff_ptr3           long      0

'  Start of HUB overlay ------------------------------------------------------------------------
' Some locations within the next set of values, after being init'd to zero, are then filled from spin with options
' That are transferred to and accessed by the pasm cog once started, but no longer needed in spin.
' Therefore, tx and rx buffers start here and overlay the hub footprint of these variables.
' tx_buffers come first, 0,1,2,3, then rx buffers 0,1,2,3 by offset from "buffers"
overlay
buffers
txdata                long      0
txbits                long      0
txcnt                 long      0
txdata1               long      0
txbits1               long      0
txcnt1                long      0
txdata2               long      0
txbits2               long      0
txcnt2                long      0
txdata3               long      0
txbits3               long      0
txcnt3                long      0
rxdata                long      0
rxbits                long      0
rxcnt                 long      0
rxdata1               long      0
rxbits1               long      0
rxcnt1                long      0
rxdata2               long      0
rxbits2               long      0
rxcnt2                long      0
rxdata3               long      0
rxbits3               long      0
rxcnt3                long      0
t1                    long      0               ' this is a temporary variable used by pasm
rxmask                long      0               ' mask for the pin used for receive, zero if port not used for receive
rxmask1               long      0
rxmask2               long      0
rxmask3               long      0
txmask                long      0               ' mask for the pin used for transmit, zero if port not used for transmit
txmask1               long      0
txmask2               long      0
txmask3               long      0
bit4_ticks            long      0             ' bit ticks for start bit, 1/4 of standard bit
bit4_ticks1           long      0
bit4_ticks2           long      0
bit4_ticks3           long      0
rxbuff_head_ptr         long      0             ' Hub address of data received, base plus offset...
rxbuff_head_ptr1        long      0             ' pasm writes to hub at this address, initialized in spin to base address
rxbuff_head_ptr2        long      0
rxbuff_head_ptr3        long      0
txbuff_tail_ptr         long      0             ' Hub address of data tranmitted, base plus offset...
txbuff_tail_ptr1        long      0             ' pasm reads from hub at this address, initialized in spin to base address
txbuff_tail_ptr2        long      0
txbuff_tail_ptr3        long      0
rx_head_ptr             long      0             ' pointer to the hub address of where the head and tail offset pointers are stored
rx_head_ptr1            long      0             ' these pointers are initialized in spin but then used only by pasm
rx_head_ptr2            long      0             ' the pasm cog has to know where in the hub to find those offsets.
rx_head_ptr3            long      0
rx_tail_ptr             long      0
rx_tail_ptr1            long      0
rx_tail_ptr2            long      0
rx_tail_ptr3            long      0
tx_head_ptr             long      0
tx_head_ptr1            long      0
tx_head_ptr2            long      0
tx_head_ptr3            long      0
tx_tail_ptr             long      0
tx_tail_ptr1            long      0
tx_tail_ptr2            long      0
tx_tail_ptr3            long      0
version                 long      1             ' for 8  ports, change this variable to 2 and save the object under a different name.
       '' ----------- End of the  object memory zeroed from startfill to endfill in the init/stop method ------
endfill
      FIT
'' The above is all of the necessary code that must fit in the cog
'' The following are extra bytes if necessary to provide the required rx and tx buffers.
'' the number required is computed from the aggregate buffer size declared, minus the above initialized but recycled variables.

extra                   byte    0 [RXTX_BUFSIZE - (RXTX_BUFSIZE <# (@extra - @overlay))]

{{
;==============================================================================================================================;
|                                                   TERMS OF USE: MIT License                                                  |
|------------------------------------------------------------------------------------------------------------------------------|
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    |
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
'=============================================================================================================================='
}}
